/*
	Copyright (C) 2011  Andrew Cotter

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


/**
	\internal
	\file clipped_voronoi_fast.hh
	\brief ClippedVoronoi class (faster but uglier version)
*/




#ifndef __CLIPPED_VORONOI_FAST_HH__
#define __CLIPPED_VORONOI_FAST_HH__

#ifdef __cplusplus




#include "vector.hh"
#include "accuracy.hh"
#include "client_data.hh"
#include "choose.hh"
#include "pool.hh"
#include "helpers.hh"

#include <boost/array.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/assert.hpp>
#include <boost/static_assert.hpp>

#include <vector>
#include <algorithm>
#include <limits>
#include <stdexcept>
#include <cmath>




//============================================================================
//    ExactCastTo helper template
//============================================================================


namespace {


template< typename t_Type >
struct ExactCastTo;


template<>
struct ExactCastTo< Accuracy::Exact > {

	static inline Accuracy::Exact const& Get( Accuracy::Exact const& exact ) {

		return exact;
	}
};


template<>
struct ExactCastTo< Accuracy::Approximate > {

	static inline Accuracy::Approximate const& Get( Accuracy::Exact const& exact ) {

		return exact.GetApproximate();
	}
};


template<>
struct ExactCastTo< mpq_class > {

	static inline mpq_class const& Get( Accuracy::Exact const& exact ) {

		return exact.GetExact();
	}
};


}    // anonymous namespace




//============================================================================
//    Intersection helper template
//============================================================================


namespace {


template< unsigned int t_Dimension >
struct Intersection;


template<>
struct Intersection< 2 > {

	template< typename t_Type >
	static inline void Solve(
		boost::array< boost::array< Accuracy::Exact, 2 > const*, 2 > const& alphasArray,
		boost::array< Accuracy::Exact, 2 > const intercepts,
		boost::array< t_Type, 2 >& resultNumerator,
		t_Type& resultDenominator
	)
	{
		resultDenominator = (
			ExactCastTo< t_Type >::Get( ( *alphasArray[ 0 ] )[ 0 ] ) * ExactCastTo< t_Type >::Get( ( *alphasArray[ 1 ] )[ 1 ] ) -
			ExactCastTo< t_Type >::Get( ( *alphasArray[ 0 ] )[ 1 ] ) * ExactCastTo< t_Type >::Get( ( *alphasArray[ 1 ] )[ 0 ] )
		);

		resultNumerator[ 0 ] = (
			ExactCastTo< t_Type >::Get( ( *alphasArray[ 1 ] )[ 1 ] ) * ExactCastTo< t_Type >::Get( intercepts[ 0 ] ) -
			ExactCastTo< t_Type >::Get( ( *alphasArray[ 0 ] )[ 1 ] ) * ExactCastTo< t_Type >::Get( intercepts[ 1 ] )
		);
		resultNumerator[ 1 ] = (
			ExactCastTo< t_Type >::Get( ( *alphasArray[ 0 ] )[ 0 ] ) * ExactCastTo< t_Type >::Get( intercepts[ 1 ] ) -
			ExactCastTo< t_Type >::Get( ( *alphasArray[ 1 ] )[ 0 ] ) * ExactCastTo< t_Type >::Get( intercepts[ 0 ] )
		);
	}
};


template<>
struct Intersection< 3 > {

	template< typename t_Type >
	static inline void Solve(
		boost::array< boost::array< Accuracy::Exact, 3 > const*, 3 > const& alphasArray,
		boost::array< Accuracy::Exact, 3 > const intercepts,
		boost::array< t_Type, 3 >& resultNumerator,
		t_Type& resultDenominator
	)
	{
		t_Type const determinant00 = (
			ExactCastTo< t_Type >::Get( ( *alphasArray[ 1 ] )[ 1 ] ) * ExactCastTo< t_Type >::Get( ( *alphasArray[ 2 ] )[ 2 ] ) -
			ExactCastTo< t_Type >::Get( ( *alphasArray[ 1 ] )[ 2 ] ) * ExactCastTo< t_Type >::Get( ( *alphasArray[ 2 ] )[ 1 ] )
		);
		t_Type const determinant10 = (
			ExactCastTo< t_Type >::Get( ( *alphasArray[ 0 ] )[ 1 ] ) * ExactCastTo< t_Type >::Get( ( *alphasArray[ 2 ] )[ 2 ] ) -
			ExactCastTo< t_Type >::Get( ( *alphasArray[ 0 ] )[ 2 ] ) * ExactCastTo< t_Type >::Get( ( *alphasArray[ 2 ] )[ 1 ] )
		);
		t_Type const determinant20 = (
			ExactCastTo< t_Type >::Get( ( *alphasArray[ 0 ] )[ 1 ] ) * ExactCastTo< t_Type >::Get( ( *alphasArray[ 1 ] )[ 2 ] ) -
			ExactCastTo< t_Type >::Get( ( *alphasArray[ 0 ] )[ 2 ] ) * ExactCastTo< t_Type >::Get( ( *alphasArray[ 1 ] )[ 1 ] )
		);

		resultDenominator = (
			determinant00 * ExactCastTo< t_Type >::Get( ( *alphasArray[ 0 ] )[ 0 ] ) -
			determinant10 * ExactCastTo< t_Type >::Get( ( *alphasArray[ 1 ] )[ 0 ] ) +
			determinant20 * ExactCastTo< t_Type >::Get( ( *alphasArray[ 2 ] )[ 0 ] )
		);

		t_Type const determinant01 = (
			ExactCastTo< t_Type >::Get( ( *alphasArray[ 1 ] )[ 0 ] ) * ExactCastTo< t_Type >::Get( ( *alphasArray[ 2 ] )[ 2 ] ) -
			ExactCastTo< t_Type >::Get( ( *alphasArray[ 1 ] )[ 2 ] ) * ExactCastTo< t_Type >::Get( ( *alphasArray[ 2 ] )[ 0 ] )
		);
		t_Type const determinant11 = (
			ExactCastTo< t_Type >::Get( ( *alphasArray[ 0 ] )[ 0 ] ) * ExactCastTo< t_Type >::Get( ( *alphasArray[ 2 ] )[ 2 ] ) -
			ExactCastTo< t_Type >::Get( ( *alphasArray[ 0 ] )[ 2 ] ) * ExactCastTo< t_Type >::Get( ( *alphasArray[ 2 ] )[ 0 ] )
		);
		t_Type const determinant21 = (
			ExactCastTo< t_Type >::Get( ( *alphasArray[ 0 ] )[ 0 ] ) * ExactCastTo< t_Type >::Get( ( *alphasArray[ 1 ] )[ 2 ] ) -
			ExactCastTo< t_Type >::Get( ( *alphasArray[ 0 ] )[ 2 ] ) * ExactCastTo< t_Type >::Get( ( *alphasArray[ 1 ] )[ 0 ] )
		);

		t_Type const determinant02 = (
			ExactCastTo< t_Type >::Get( ( *alphasArray[ 1 ] )[ 0 ] ) * ExactCastTo< t_Type >::Get( ( *alphasArray[ 2 ] )[ 1 ] ) -
			ExactCastTo< t_Type >::Get( ( *alphasArray[ 1 ] )[ 1 ] ) * ExactCastTo< t_Type >::Get( ( *alphasArray[ 2 ] )[ 0 ] )
		);
		t_Type const determinant12 = (
			ExactCastTo< t_Type >::Get( ( *alphasArray[ 0 ] )[ 0 ] ) * ExactCastTo< t_Type >::Get( ( *alphasArray[ 2 ] )[ 1 ] ) -
			ExactCastTo< t_Type >::Get( ( *alphasArray[ 0 ] )[ 1 ] ) * ExactCastTo< t_Type >::Get( ( *alphasArray[ 2 ] )[ 0 ] )
		);
		t_Type const determinant22 = (
			ExactCastTo< t_Type >::Get( ( *alphasArray[ 0 ] )[ 0 ] ) * ExactCastTo< t_Type >::Get( ( *alphasArray[ 1 ] )[ 1 ] ) -
			ExactCastTo< t_Type >::Get( ( *alphasArray[ 0 ] )[ 1 ] ) * ExactCastTo< t_Type >::Get( ( *alphasArray[ 1 ] )[ 0 ] )
		);

		resultNumerator[ 0 ] = (
			determinant00 * ExactCastTo< t_Type >::Get( intercepts[ 0 ] ) -
			determinant10 * ExactCastTo< t_Type >::Get( intercepts[ 1 ] ) +
			determinant20 * ExactCastTo< t_Type >::Get( intercepts[ 2 ] )
		);
		resultNumerator[ 1 ] = (
			-determinant01 * ExactCastTo< t_Type >::Get( intercepts[ 0 ] ) +
			 determinant11 * ExactCastTo< t_Type >::Get( intercepts[ 1 ] ) -
			 determinant21 * ExactCastTo< t_Type >::Get( intercepts[ 2 ] )
		);
		resultNumerator[ 2 ] = (
			determinant02 * ExactCastTo< t_Type >::Get( intercepts[ 0 ] ) -
			determinant12 * ExactCastTo< t_Type >::Get( intercepts[ 1 ] ) +
			determinant22 * ExactCastTo< t_Type >::Get( intercepts[ 2 ] )
		);
	}
};


// **TODO: specializations for 4x4 and higher-dimensional matrices


}    // anonymous namespace




//============================================================================
//    ClippedVoronoi class
//============================================================================


template< unsigned int t_Dimension, typename t_Type = double, typename t_ClientDataType = void >
struct ClippedVoronoi {

	BOOST_STATIC_ASSERT( t_Dimension > 1 );


	enum { DIMENSION = t_Dimension };
	typedef t_Type Type;


	// returned by Test? methods
	enum Test {
		TEST_UNKNOWN = -1,
		TEST_OUTSIDE = 0,
		TEST_INSIDE = 1
	};


private:

	struct Hyperplane;
	struct Vertex;


public:

	inline ClippedVoronoi(
		PoolList& poolList,
		boost::array< t_Type, t_Dimension > const& center,
		Accuracy::Exact const& nearestNeighborDistanceSquared,
		double const rho,
		double const beta
	);

	inline ~ClippedVoronoi();


	inline boost::array< t_Type, t_Dimension > const& GetCenter() const;

	inline bool const IsWellSpaced() const;


	//====  Neighbor accessors  ====

	inline unsigned int const GetNeighbors() const;

	inline ClientData< t_ClientDataType >* const GetClientData( unsigned int const index );
	inline ClientData< t_ClientDataType > const* const GetClientData( unsigned int const index ) const;


	//====  Test if a neighbor could be at a certain distance, by checking the bounding spheres  ====

	inline Test const ApproximateTestDistanceSquared( Accuracy::Approximate const& distanceSquared ) const;


	//====  Test if a box contains a neighbor, by checking the Voronoi vertices  ====

	inline Test const ApproximateTestBox(
		boost::array< t_Type, t_Dimension > const& lower,
		boost::array< t_Type, t_Dimension > const& upper
	) const;


	//====  Insert a new neighbor point  ====

	ClientData< t_ClientDataType >* const InsertNeighbor( boost::array< t_Type, t_Dimension > const& point );


	//====  Check/find a Steiner point  ====

	inline bool const IsSteiner( boost::array< t_Type, t_Dimension > const& point ) const;

	// **NOTE: returns NULL if no Steiner point exists
	boost::shared_ptr< boost::array< t_Type, t_Dimension > > const FindSteiner( double const lambda = 0.5 ) const;


private:

	inline void RecalculateOuterRadiusSquared();
	inline void RecalculateInnerRadiusSquared();

	inline void CalculateExactOuterRadiusSquared() const;
	inline void CalculateExact( Vertex* const pVertex ) const;

	inline Test const TestPoint( boost::array< t_Type, t_Dimension > const& point ) const;


	struct Hyperplane : public ClientData< t_ClientDataType > {

		boost::array< Accuracy::Exact, t_Dimension > alphas;
		Accuracy::Exact intercept;
		std::vector< Vertex* > vertices;
		bool flag;
	};

	struct Vertex {

		boost::array< Accuracy::Approximate, t_Dimension > approximatePoint;
		boost::array< mpq_class, t_Dimension >* pExactPoint;

		boost::array< Hyperplane*, t_Dimension > hyperplanes;
	};


	boost::array< t_Type, t_Dimension > m_center;    // everything stored by this structure is centered around this point

	Accuracy::Exact m_innerBoundSquared;    // rho*NN
	Accuracy::Exact m_outerBoundSquared;    // beta*NN

	Accuracy::Approximate m_approximateOuterRadiusSquared;    // maximum distance to a vertex
	mpq_class* m_pExactOuterRadiusSquared;

	Accuracy::Exact m_innerRadiusSquared;    // minimum distance to a hyperplane

	std::vector< Hyperplane* > m_hyperplanes;    // this vector owns the hyperplanes
	std::vector< Vertex*     > m_vertices;       // this vector owns the vertices

	Pool* m_pHyperplanePool;
	Pool* m_pVertexPool;


	// private and unimplemented
	inline ClippedVoronoi( ClippedVoronoi const& );
	inline ClippedVoronoi const& operator=( ClippedVoronoi const& );
};




//============================================================================
//    ClippedVoronoi inline methods
//============================================================================


template< unsigned int t_Dimension, typename t_Type, typename t_ClientDataType >
ClippedVoronoi< t_Dimension, t_Type, t_ClientDataType >::ClippedVoronoi(
	PoolList& poolList,
	boost::array< t_Type, t_Dimension > const& center,
	Accuracy::Exact const& nearestNeighborDistanceSquared,
	double const rho,
	double const beta
) :
	m_center( center ),
	m_innerBoundSquared( nearestNeighborDistanceSquared * Square( Accuracy::ExactCast( rho  ) ) ),
	m_outerBoundSquared( nearestNeighborDistanceSquared * Square( Accuracy::ExactCast( beta ) ) ),
	m_pExactOuterRadiusSquared( NULL ),
	m_pHyperplanePool( poolList.GetPool( sizeof( Hyperplane ) ) ),
	m_pVertexPool( poolList.GetPool( sizeof( Vertex ) ) )
{
	BOOST_ASSERT( rho  >  1       );
	BOOST_ASSERT( beta >= rho     );
	BOOST_ASSERT( beta <  2 * rho );

	//====  Start off with a box of width 2*sqrt(m_outerBoundSquared), clipped to [0,1]  ====

	boost::array< Accuracy::Exact, t_Dimension > lower;
	boost::array< Accuracy::Exact, t_Dimension > upper;
	{	// epsilon is how far outside of the [0,1] bounding box our "virtual" bounding box will extend
		Accuracy::Exact bound = Accuracy::ExactCast( std::sqrt( m_outerBoundSquared.GetApproximate().GetUpperBound() ) );

		for ( unsigned int ii = 0; ii < t_Dimension; ++ii ) {

			lower[ ii ] = Accuracy::ExactCast( m_center[ ii ] );
			lower[ ii ] = std::min( bound, lower[ ii ] );
			BOOST_ASSERT( lower[ ii ] >= Accuracy::ExactCast( 0.0 ) );

			upper[ ii ] = Accuracy::ExactCast( 1.0 ) - Accuracy::ExactCast( m_center[ ii ] );
			upper[ ii ] = std::min( bound, upper[ ii ] );
			BOOST_ASSERT( upper[ ii ] >= Accuracy::ExactCast( 0.0 ) );
		}
	}

	for ( unsigned int ii = 0; ii < t_Dimension; ++ii ) {

		{	Hyperplane* const pHyperplane = static_cast< Hyperplane* >( m_pHyperplanePool->Allocate() );
			new ( pHyperplane ) Hyperplane;
			std::fill( pHyperplane->alphas.begin(), pHyperplane->alphas.end(), Accuracy::ExactCast( 0.0 ) );
			pHyperplane->alphas[ ii ] = Accuracy::ExactCast( -1.0 );
			pHyperplane->intercept = lower[ ii ];
			m_hyperplanes.push_back( pHyperplane );
		}

		{	Hyperplane* const pHyperplane = static_cast< Hyperplane* >( m_pHyperplanePool->Allocate() );
			new ( pHyperplane ) Hyperplane;
			std::fill( pHyperplane->alphas.begin(), pHyperplane->alphas.end(), Accuracy::ExactCast( 0.0 ) );
			pHyperplane->alphas[ ii ] = Accuracy::ExactCast( 1.0 );
			pHyperplane->intercept = upper[ ii ];
			m_hyperplanes.push_back( pHyperplane );
		}
	}

	for ( unsigned int ii = 0; ii < ( 1ull << t_Dimension ); ++ii ) {

		Vertex* const pVertex = static_cast< Vertex* >( m_pVertexPool->Allocate() );
		new ( pVertex ) Vertex;
		for ( unsigned int jj = 0; jj < t_Dimension; ++jj ) {

			if ( ( ii & ( 1ull << jj ) ) == 0 ) {

				pVertex->approximatePoint[ jj ] = -lower[ jj ].GetApproximate();
				pVertex->pExactPoint = NULL;
				pVertex->hyperplanes[ jj ] = m_hyperplanes[ jj * 2 ];
				m_hyperplanes[ jj * 2 ]->vertices.push_back( pVertex );
			}
			else {

				pVertex->approximatePoint[ jj ] = upper[ jj ].GetApproximate();
				pVertex->pExactPoint = NULL;
				pVertex->hyperplanes[ jj ] = m_hyperplanes[ jj * 2 + 1 ];
				m_hyperplanes[ jj * 2 + 1 ]->vertices.push_back( pVertex );
			}
		}
		m_vertices.push_back( pVertex );
	}

	RecalculateOuterRadiusSquared();
	RecalculateInnerRadiusSquared();
}


template< unsigned int t_Dimension, typename t_Type, typename t_ClientDataType >
ClippedVoronoi< t_Dimension, t_Type, t_ClientDataType >::~ClippedVoronoi() {

	{	typename std::vector< Hyperplane* >::const_iterator ii    = m_hyperplanes.begin();
		typename std::vector< Hyperplane* >::const_iterator iiEnd = m_hyperplanes.end();
		for ( ; ii != iiEnd; ++ii ) {

			( *ii )->~Hyperplane();
			m_pHyperplanePool->Free( *ii );
		}
	}

	{	typename std::vector< Vertex* >::const_iterator ii    = m_vertices.begin();
		typename std::vector< Vertex* >::const_iterator iiEnd = m_vertices.end();
		for ( ; ii != iiEnd; ++ii ) {

			if ( ( *ii )->pExactPoint != NULL )
				delete ( *ii )->pExactPoint;
			( *ii )->~Vertex();
			m_pVertexPool->Free( *ii );
		}
	}
}


template< unsigned int t_Dimension, typename t_Type, typename t_ClientDataType >
boost::array< t_Type, t_Dimension > const& ClippedVoronoi< t_Dimension, t_Type, t_ClientDataType >::GetCenter() const {

	return m_center;
}


template< unsigned int t_Dimension, typename t_Type, typename t_ClientDataType >
bool const ClippedVoronoi< t_Dimension, t_Type, t_ClientDataType >::IsWellSpaced() const {

	bool result = false;
	switch( Accuracy::Approximate::Compare( m_approximateOuterRadiusSquared, m_innerBoundSquared.GetApproximate() ) ) {

		case -1: result = true;  break;
		case  1: result = false; break;

		case 0: {

			CalculateExactOuterRadiusSquared();
			BOOST_ASSERT( m_pExactOuterRadiusSquared != NULL );
			result = ( *m_pExactOuterRadiusSquared <= m_innerBoundSquared.GetExact() );
			break;
		}

		default: BOOST_ASSERT( false ); break;
	}

	return result;
}


template< unsigned int t_Dimension, typename t_Type, typename t_ClientDataType >
unsigned int const ClippedVoronoi< t_Dimension, t_Type, t_ClientDataType >::GetNeighbors() const {

	return m_hyperplanes.size();
}


template< unsigned int t_Dimension, typename t_Type, typename t_ClientDataType >
ClientData< t_ClientDataType >* const ClippedVoronoi< t_Dimension, t_Type, t_ClientDataType >::GetClientData( unsigned int const index ) {

	BOOST_ASSERT( index < m_hyperplanes.size() );
	return m_hyperplanes[ index ];
}


template< unsigned int t_Dimension, typename t_Type, typename t_ClientDataType >
ClientData< t_ClientDataType > const* const ClippedVoronoi< t_Dimension, t_Type, t_ClientDataType >::GetClientData( unsigned int const index ) const {

	BOOST_ASSERT( index < m_hyperplanes.size() );
	return m_hyperplanes[ index ];
}


template< unsigned int t_Dimension, typename t_Type, typename t_ClientDataType >
typename ClippedVoronoi< t_Dimension, t_Type, t_ClientDataType >::Test const ClippedVoronoi< t_Dimension, t_Type, t_ClientDataType >::ApproximateTestDistanceSquared( Accuracy::Approximate const& distanceSquared ) const {

	Test result = TEST_OUTSIDE;

	if (
		( Accuracy::Approximate::Compare( distanceSquared, m_outerBoundSquared.GetApproximate() * 4 ) <= 0 ) &&
		( Accuracy::Approximate::Compare( distanceSquared, m_approximateOuterRadiusSquared      * 4 ) <= 0 )
	)
	{
		if ( Accuracy::Approximate::Compare( distanceSquared, m_innerRadiusSquared.GetApproximate() * 4 ) < 0 )
			result = TEST_INSIDE;
		else
			result = TEST_UNKNOWN;
	}

	return result;
}


template< unsigned int t_Dimension, typename t_Type, typename t_ClientDataType >
typename ClippedVoronoi< t_Dimension, t_Type, t_ClientDataType >::Test const ClippedVoronoi< t_Dimension, t_Type, t_ClientDataType >::ApproximateTestBox(
	boost::array< t_Type, t_Dimension > const& lower,
	boost::array< t_Type, t_Dimension > const& upper
) const
{
	Test result = TEST_OUTSIDE;

	// check if any vertices are closer to the box than to the center, using exact math
	typename std::vector< Vertex* >::const_iterator ii    = m_vertices.begin();
	typename std::vector< Vertex* >::const_iterator iiEnd = m_vertices.end();
	for ( ; ii != iiEnd; ++ii ) {

		Accuracy::Approximate normSquared( EuclideanNormSquared( ( *ii )->approximatePoint ) );

		Accuracy::Approximate distanceSquared( 0 );
		for ( unsigned int jj = 0; jj < t_Dimension; ++jj ) {

			/*
				**FIXME: We currently err on the side of making
				distanceSquared too small. As a result, we can never know
				for sure if we should return TEST_INSIDE
			*/
			Accuracy::Approximate const coordinate( m_center[ jj ] + ( *ii )->approximatePoint[ jj ] );
			if ( Accuracy::Approximate::Compare( coordinate, lower[ jj ] ) < 0 )
				distanceSquared += Square( lower[ jj ] - coordinate );
			else if ( Accuracy::Approximate::Compare( coordinate, upper[ jj ] ) > 0 )
				distanceSquared += Square( upper[ jj ] - coordinate );
		}

		if ( Accuracy::Approximate::Compare( distanceSquared, normSquared ) <= 0 ) {

			result = TEST_UNKNOWN;
			break;
		}
	}

	return result;
}


template< unsigned int t_Dimension, typename t_Type, typename t_ClientDataType >
bool const ClippedVoronoi< t_Dimension, t_Type, t_ClientDataType >::IsSteiner( boost::array< t_Type, t_Dimension > const& point ) const {

	Accuracy::Exact const distanceSquared = ExactEuclideanDistanceSquared( m_center, point );

	return(
		( Accuracy::Approximate::Compare( distanceSquared.GetApproximate(), m_approximateOuterRadiusSquared ) <= 0 ) &&
		( distanceSquared >= m_innerBoundSquared ) &&
		( distanceSquared <= m_outerBoundSquared ) &&
		( TestPoint( point ) == TEST_INSIDE )
	);
}


template< unsigned int t_Dimension, typename t_Type, typename t_ClientDataType >
void ClippedVoronoi< t_Dimension, t_Type, t_ClientDataType >::RecalculateOuterRadiusSquared() {

	if ( m_pExactOuterRadiusSquared != NULL ) {

		delete m_pExactOuterRadiusSquared;
		m_pExactOuterRadiusSquared = NULL;
	}

	// m_approximateOuterRadiusSquared/*m_pExactOuterRadiusSquared is maximum distance to a vertex
	{	m_approximateOuterRadiusSquared = Accuracy::Approximate( 0.0 );

		typename std::vector< Vertex* >::const_iterator ii    = m_vertices.begin();
		typename std::vector< Vertex* >::const_iterator iiEnd = m_vertices.end();
		for ( ; ii != iiEnd; ++ii ) {

			Accuracy::Approximate const normSquared( EuclideanNormSquared( ( *ii )->approximatePoint ) );
			m_approximateOuterRadiusSquared = std::max( m_approximateOuterRadiusSquared, normSquared );
		}
	}
}


template< unsigned int t_Dimension, typename t_Type, typename t_ClientDataType >
void ClippedVoronoi< t_Dimension, t_Type, t_ClientDataType >::RecalculateInnerRadiusSquared() {

	// m_innerRadius is minimum distance to a hyperplane
	{	m_innerRadiusSquared = m_outerBoundSquared;

		typename std::vector< Hyperplane* >::iterator ii    = m_hyperplanes.begin();
		typename std::vector< Hyperplane* >::iterator iiEnd = m_hyperplanes.end();
		for ( ; ii != iiEnd; ++ii ) {

			Accuracy::Exact const normSquared = Square( ( *ii )->intercept ) / EuclideanNormSquared( ( *ii )->alphas );
			m_innerRadiusSquared = std::min( m_innerRadiusSquared, normSquared );
		}
	}
	BOOST_ASSERT( m_innerRadiusSquared > Accuracy::ExactCast( 0.0 ) );
	BOOST_ASSERT( Accuracy::Approximate::Compare( m_innerRadiusSquared.GetApproximate(), m_approximateOuterRadiusSquared ) <= 0 );
}


template< unsigned int t_Dimension, typename t_Type, typename t_ClientDataType >
void ClippedVoronoi< t_Dimension, t_Type, t_ClientDataType >::CalculateExactOuterRadiusSquared() const {

	if ( m_pExactOuterRadiusSquared == NULL ) {

		const_cast< mpq_class*& >( m_pExactOuterRadiusSquared ) = new mpq_class( 0.0 );

		typename std::vector< Vertex* >::const_iterator ii    = m_vertices.begin();
		typename std::vector< Vertex* >::const_iterator iiEnd = m_vertices.end();
		for ( ; ii != iiEnd; ++ii ) {

			Accuracy::Approximate approximateNormSquared( EuclideanNormSquared( ( *ii )->approximatePoint ) );
			if ( Accuracy::Approximate::Compare( approximateNormSquared, m_approximateOuterRadiusSquared ) >= 0 ) {

				CalculateExact( *ii );
				BOOST_ASSERT( ( *ii )->pExactPoint != NULL );

				mpq_class exactNormSquared( Square( ( *( *ii )->pExactPoint )[ 0 ] ) );
				for ( unsigned int jj = 1; jj < t_Dimension; ++jj )
					exactNormSquared += Square( ( *( *ii )->pExactPoint )[ jj ] );

				if ( exactNormSquared > *m_pExactOuterRadiusSquared )
					*m_pExactOuterRadiusSquared = exactNormSquared;
			}
		}

		const_cast< Accuracy::Approximate& >( m_approximateOuterRadiusSquared ) = Accuracy::Approximate( m_pExactOuterRadiusSquared->get_d() );
	}
	BOOST_ASSERT( m_pExactOuterRadiusSquared != NULL );
	BOOST_ASSERT( *m_pExactOuterRadiusSquared > 0.0 );
}


template< unsigned int t_Dimension, typename t_Type, typename t_ClientDataType >
void ClippedVoronoi< t_Dimension, t_Type, t_ClientDataType >::CalculateExact( Vertex* const pVertex ) const {

	if ( pVertex->pExactPoint == NULL ) {

		boost::array< boost::array< Accuracy::Exact, t_Dimension > const*, t_Dimension > alphasArray;
		boost::array< Accuracy::Exact, t_Dimension > intercepts;

		for ( unsigned int ii = 0; ii < t_Dimension; ++ii ) {

			alphasArray[ ii ] = &pVertex->hyperplanes[ ii ]->alphas;
			intercepts[  ii ] =  pVertex->hyperplanes[ ii ]->intercept;
		}

		pVertex->pExactPoint = new boost::array< mpq_class, t_Dimension >;
		mpq_class denominator;
		Intersection< t_Dimension >::Solve( alphasArray, intercepts, *pVertex->pExactPoint, denominator );
		BOOST_ASSERT( denominator != 0.0 );

		for ( unsigned int ii = 0; ii < t_Dimension; ++ii ) {

			( *pVertex->pExactPoint )[ ii ] = ( *pVertex->pExactPoint )[ ii ] / denominator;
			pVertex->approximatePoint[ ii ] = Accuracy::Approximate( ( *pVertex->pExactPoint )[ ii ].get_d() );
		}
	}
	BOOST_ASSERT( pVertex->pExactPoint != NULL );
}


template< unsigned int t_Dimension, typename t_Type, typename t_ClientDataType >
typename ClippedVoronoi< t_Dimension, t_Type, t_ClientDataType >::Test const ClippedVoronoi< t_Dimension, t_Type, t_ClientDataType >::TestPoint(
	boost::array< t_Type, t_Dimension > const& point
) const
{
	Test result = TEST_INSIDE;

	// check if the point is on the wrong side of any hyperplanes, using exact math
	typename std::vector< Hyperplane* >::const_iterator ii    = m_hyperplanes.begin();
	typename std::vector< Hyperplane* >::const_iterator iiEnd = m_hyperplanes.end();
	for ( ; ii != iiEnd; ++ii ) {

		if ( ExactDotProduct( point, ( *ii )->alphas ) - ExactDotProduct( m_center, ( *ii )->alphas ) > ( *ii )->intercept ) {

			result = TEST_OUTSIDE;
			break;
		}
	}

	return result;
}




//============================================================================
//    ClippedVoronoi methods
//============================================================================


template< unsigned int t_Dimension, typename t_Type, typename t_ClientDataType >
ClientData< t_ClientDataType >* const ClippedVoronoi< t_Dimension, t_Type, t_ClientDataType >::InsertNeighbor( boost::array< t_Type, t_Dimension > const& point ) {

	ClientData< t_ClientDataType >* pResult = NULL;

	boost::array< Accuracy::Exact, t_Dimension > alphas;
	for ( unsigned int ii = 0; ii < t_Dimension; ++ii )
		alphas[ ii ] = ( Accuracy::ExactCast( point[ ii ] ) - Accuracy::ExactCast( m_center[ ii ] ) ) * Accuracy::ExactCast( 0.5 );
	Accuracy::Exact const intercept = EuclideanNormSquared( alphas );
	BOOST_ASSERT( intercept > Accuracy::ExactCast( 0.0 ) );

	if (
		( Accuracy::Approximate::Compare( intercept.GetApproximate(), m_approximateOuterRadiusSquared ) <= 0 ) &&
		( intercept <= m_outerBoundSquared  )
	)
	{
		BOOST_ASSERT( m_hyperplanes.size() > t_Dimension );

		bool outerRadiusChanged = false;
		bool innerRadiusChanged = false;


		//====  remove vertices which are on the wrong side of the new hyperplane  ====

		{	typename std::vector< Hyperplane* >::const_iterator ii    = m_hyperplanes.begin();
			typename std::vector< Hyperplane* >::const_iterator iiEnd = m_hyperplanes.end();
			for ( ; ii != iiEnd; ++ii )
				( *ii )->flag = false;
		}

		bool insert = false;
		{	int iiEnd = m_vertices.size();
			for ( int ii = 0; ii < iiEnd; ) {

				int comparison = 0;
				{	Accuracy::Approximate approximateDotProduct( m_vertices[ ii ]->approximatePoint[ 0 ] * alphas[ 0 ].GetApproximate() );
					for ( unsigned int jj = 1; jj < t_Dimension; ++jj )
						approximateDotProduct += m_vertices[ ii ]->approximatePoint[ jj ] * alphas[ jj ].GetApproximate();

					comparison = Accuracy::Approximate::Compare( approximateDotProduct, intercept.GetApproximate() );
					if ( comparison == 0 ) {

						CalculateExact( m_vertices[ ii ] );
						BOOST_ASSERT( m_vertices[ ii ]->pExactPoint != NULL );

						mpq_class exactDotProduct( ( *m_vertices[ ii ]->pExactPoint )[ 0 ] * alphas[ 0 ].GetExact() );
						for ( unsigned int jj = 1; jj < t_Dimension; ++jj )
							exactDotProduct += ( *m_vertices[ ii ]->pExactPoint )[ jj ] * alphas[ jj ].GetExact();

						if ( exactDotProduct >= intercept.GetExact() ) {

							if ( exactDotProduct == intercept.GetExact() )
								comparison = 0;
							else
								comparison = 1;
						}
						else
							comparison = -1;
					}
				}

				if ( comparison >= 0 ) {

					if ( comparison > 0 ) {    // the new hyperplane cuts off the vertex

						for ( unsigned int jj = 0; jj < t_Dimension; ++jj ) {

							// set the flag for any hyperplanes involved in a removed vertex
							Hyperplane* const pHyperplane = m_vertices[ ii ]->hyperplanes[ jj ];
							pHyperplane->flag = true;

							std::vector< Vertex* >& vertices = pHyperplane->vertices;

							typename std::vector< Vertex* >::iterator kk = vertices.begin();
							for ( ; *kk != m_vertices[ ii ]; ++kk )
								BOOST_ASSERT( kk != vertices.end() );

							vertices.erase( kk );
						}

						if ( ! outerRadiusChanged ) {

							// check if this point could have been at m_outerRadiusSquared
							Accuracy::Approximate approximateNormSquared( Square( m_vertices[ ii ]->approximatePoint[ 0 ] ) );
							for ( unsigned int kk = 1; kk < t_Dimension; ++kk )
								approximateNormSquared += Square( m_vertices[ ii ]->approximatePoint[ kk ] );
							if ( Accuracy::Approximate::Compare( approximateNormSquared, m_approximateOuterRadiusSquared ) >= 0 )
								outerRadiusChanged = true;
						}

						typename std::vector< Vertex* >::iterator jj = m_vertices.begin() + ii;
						if ( ( *jj )->pExactPoint != NULL )
							delete ( *jj )->pExactPoint;
						( *jj )->~Vertex();
						m_pVertexPool->Free( *jj );
						m_vertices.erase( jj );

						insert = true;

						--iiEnd;
					}
					else {    // the new hyperplane intersects with the vertex

						for ( unsigned int jj = 0; jj < t_Dimension; ++jj )
							m_vertices[ ii ]->hyperplanes[ jj ]->flag = true;

						++ii;
					}
				}
				else
					++ii;
			}
		}

		if ( insert ) {

			if ( ! innerRadiusChanged ) {

				// check if this hyperplane could be closer than m_innerRadiusSquared
				Accuracy::Approximate approximateNormSquared( Square( alphas[ 0 ].GetApproximate() ) );
				for ( unsigned int ii = 1; ii < t_Dimension; ++ii )
					approximateNormSquared += Square( alphas[ ii ].GetApproximate() );
				if ( Accuracy::Approximate::Compare( Square( intercept.GetApproximate() ), m_innerRadiusSquared.GetApproximate() * approximateNormSquared ) <= 0 )
					innerRadiusChanged = true;
			}


			//====  insert the new hyperplane  ====

			Hyperplane* const pHyperplane = static_cast< Hyperplane* >( m_pHyperplanePool->Allocate() );
			new ( pHyperplane ) Hyperplane;
			pHyperplane->alphas    = alphas;
			pHyperplane->intercept = intercept;
			m_hyperplanes.push_back( pHyperplane );

			pResult = pHyperplane;

			boost::array< boost::array< Accuracy::Exact, t_Dimension > const*, t_Dimension > alphasArray;
			boost::array< Accuracy::Exact, t_Dimension > intercepts;
			alphasArray[ 0 ] = &alphas;
			intercepts[  0 ] = intercept;

			for ( Choose choose( m_hyperplanes.size() - 1, t_Dimension - 1 ); ! choose.Done(); ++choose ) {

				// we only need to find intersections with hyperplanes which were involved in a removed vertex
				bool check = true;
				for ( unsigned int ii = 0; ii < t_Dimension - 1; ++ii ) {

					if ( ! m_hyperplanes[ choose[ ii ] ]->flag ) {

						check = false;
						break;
					}
				}

				/*
					in debug build, we'll try to insert the vertex, and
					assert the correctness of the check flag
				*/
#ifdef NDEBUG
				if ( check ) {
#endif    // NDEBUG

					for ( unsigned int ii = 0; ii < t_Dimension - 1; ++ii ) {

						alphasArray[ ii + 1 ] = &m_hyperplanes[ choose[ ii ] ]->alphas;
						intercepts[  ii + 1 ] =  m_hyperplanes[ choose[ ii ] ]->intercept;
					}

					boost::array< Accuracy::Approximate, t_Dimension > approximateVertexNumerator;
					Accuracy::Approximate approximateVertexDenominator;
					Intersection< t_Dimension >::Solve( alphasArray, intercepts, approximateVertexNumerator, approximateVertexDenominator );
					// **NOTE: approximateVertexDenominator may be negative, which could screw up the directions of comparisons

					// **NOTE: m_innerRadiusSquared might be invalid, since we could be inserting multiple points
					Accuracy::Approximate const approximateNumeratorNormSquared = EuclideanNormSquared( approximateVertexNumerator );
					if ( Accuracy::Approximate::Compare( approximateNumeratorNormSquared, Accuracy::Approximate( 1.01 ) * m_approximateOuterRadiusSquared * Square( approximateVertexDenominator ) ) < 0 ) {

						boost::array< Accuracy::Approximate, t_Dimension > approximateVertex;
						for ( unsigned int ii = 0; ii < t_Dimension; ++ii )
							approximateVertex[ ii ] = approximateVertexNumerator[ ii ] / approximateVertexDenominator;

						int comparison = -1;
						{	unsigned int ii = 0;
							for ( unsigned int jj = 0; comparison < 1; ++jj ) {

								while ( ( ii < t_Dimension - 1 ) && ( jj == choose[ ii ] ) ) {

									++ii;
									++jj;
								}
								if ( jj >= m_hyperplanes.size() - 1 )
									break;

								Accuracy::Approximate dotProduct( approximateVertex[ 0 ] * m_hyperplanes[ jj ]->alphas[ 0 ].GetApproximate() );
								for ( unsigned int kk = 1; kk < t_Dimension; ++ kk )
									dotProduct += approximateVertex[ kk ] * m_hyperplanes[ jj ]->alphas[ kk ].GetApproximate();

								switch( Accuracy::Approximate::Compare( dotProduct, m_hyperplanes[ jj ]->intercept.GetApproximate() ) ) {

									case -1: break;
									case  0: comparison = 0; break;
									case  1: comparison = 1; break;
									default: BOOST_ASSERT( false ); break;
								}
							}
						}

						if ( comparison < 1 ) {

							if ( comparison == 0 ) {

								boost::array< mpq_class, t_Dimension > exactVertexNumerator;
								mpq_class exactVertexDenominator;
								Intersection< t_Dimension >::Solve( alphasArray, intercepts, exactVertexNumerator, exactVertexDenominator );
								// **NOTE: exactVertexDenominator may be negative, which could screw up the directions of comparisons

								// **NOTE: m_innerRadiusSquared might be invalid, since we could be inserting multiple points
								// **NOTE: we cannot call CalculateExactOuterRadiusSquared(), since some corners have been removed
								if ( exactVertexDenominator != 0.0 ) {

									boost::array< mpq_class, t_Dimension >* const pExactVertex = new boost::array< mpq_class, t_Dimension >;
									for ( unsigned int ii = 0; ii < t_Dimension; ++ii ) {

										( *pExactVertex )[ ii ] = exactVertexNumerator[ ii ] / exactVertexDenominator;
										approximateVertex[ ii ] = Accuracy::Approximate( ( *pExactVertex )[ ii ].get_d() );
									}

									comparison = -1;
									{	unsigned int ii = 0;
										for ( unsigned int jj = 0; comparison < 1; ++jj ) {

											while ( ( ii < t_Dimension - 1 ) && ( jj == choose[ ii ] ) ) {

												++ii;
												++jj;
											}
											if ( jj >= m_hyperplanes.size() - 1 )
												break;

											Accuracy::Approximate approximateDotProduct( approximateVertex[ 0 ] * m_hyperplanes[ jj ]->alphas[ 0 ].GetApproximate() );
											for ( unsigned int kk = 1; kk < t_Dimension; ++ kk )
												approximateDotProduct += approximateVertex[ kk ] * m_hyperplanes[ jj ]->alphas[ kk ].GetApproximate();

											switch( Accuracy::Approximate::Compare( approximateDotProduct, m_hyperplanes[ jj ]->intercept.GetApproximate() ) ) {

												case -1: break;
												case  1: comparison = 1; break;

												case 0: {

													mpq_class exactDotProduct( ( *pExactVertex )[ 0 ] * m_hyperplanes[ jj ]->alphas[ 0 ].GetExact() );
													for ( unsigned int kk = 1; kk < t_Dimension; ++ kk )
														exactDotProduct += ( *pExactVertex )[ kk ] * m_hyperplanes[ jj ]->alphas[ kk ].GetExact();

													if ( exactDotProduct > m_hyperplanes[ jj ]->intercept.GetExact() )
														comparison = 1;

													break;
												}

												default: BOOST_ASSERT( false ); break;
											}
										}
									}

									if ( comparison < 1 ) {

										// we don't use the check flag in the debug build, so that we may assert it here
										BOOST_ASSERT( check );

										Vertex* const pVertex = static_cast< Vertex* >( m_pVertexPool->Allocate() );
										new ( pVertex ) Vertex;
										pVertex->approximatePoint = approximateVertex;
										pVertex->pExactPoint = pExactVertex;
										pVertex->hyperplanes[ 0 ] = pHyperplane;
										pHyperplane->vertices.push_back( pVertex );
										for ( unsigned int ii = 0; ii < t_Dimension - 1; ++ii ) {

											pVertex->hyperplanes[ ii + 1 ] = m_hyperplanes[ choose[ ii ] ];
											m_hyperplanes[ choose[ ii ] ]->vertices.push_back( pVertex );
										}
										m_vertices.push_back( pVertex );
									}
									else
										delete pExactVertex;
								}
							}
							else {

								// we don't use the check flag in the debug build, so that we may assert it here
								BOOST_ASSERT( check );

								Vertex* const pVertex = static_cast< Vertex* >( m_pVertexPool->Allocate() );
								new ( pVertex ) Vertex;
								pVertex->approximatePoint = approximateVertex;
								pVertex->pExactPoint = NULL;
								pVertex->hyperplanes[ 0 ] = pHyperplane;
								pHyperplane->vertices.push_back( pVertex );
								for ( unsigned int ii = 0; ii < t_Dimension - 1; ++ii ) {

									pVertex->hyperplanes[ ii + 1 ] = m_hyperplanes[ choose[ ii ] ];
									m_hyperplanes[ choose[ ii ] ]->vertices.push_back( pVertex );
								}
								m_vertices.push_back( pVertex );
							}
						}
					}
#ifdef NDEBUG
				}
#endif    // NDEBUG
			}

			BOOST_ASSERT( ! pHyperplane->vertices.empty() );

			//====  remove hyperplanes which are no longer involved in any vertices  ====

			{	int iiEnd = m_hyperplanes.size() - 1;
				for ( int ii = 0; ii < iiEnd; ) {

					if ( m_hyperplanes[ ii ]->vertices.empty() ) {

						typename std::vector< Hyperplane* >::iterator jj = m_hyperplanes.begin() + ii;
						( *jj )->~Hyperplane();
						m_pHyperplanePool->Free( *jj );
						m_hyperplanes.erase( jj );

						--iiEnd;
					}
					else
						++ii;
				}
			}
		}

		//====  recalculate radii  ====

		if ( outerRadiusChanged )
			RecalculateOuterRadiusSquared();
		if ( innerRadiusChanged )
			RecalculateInnerRadiusSquared();
	}

	return pResult;
}


template< unsigned int t_Dimension, typename t_Type, typename t_ClientDataType >
boost::shared_ptr< boost::array< t_Type, t_Dimension > > const ClippedVoronoi< t_Dimension, t_Type, t_ClientDataType >::FindSteiner( double const lambda ) const {

	BOOST_ASSERT( m_innerRadiusSquared > Accuracy::ExactCast( 0.0 ) );
	BOOST_ASSERT( Accuracy::Approximate::Compare( m_innerRadiusSquared.GetApproximate(), m_approximateOuterRadiusSquared ) <= 0 );

	boost::shared_ptr< boost::array< t_Type, t_Dimension > > pResult;
	bool isSteiner = false;

	if ( Accuracy::Approximate::Compare( m_approximateOuterRadiusSquared, m_innerBoundSquared.GetApproximate() ) >= 0 ) {

		typename std::vector< Vertex* >::const_iterator ii    = m_vertices.begin();
		typename std::vector< Vertex* >::const_iterator iiEnd = m_vertices.end();
		for ( ; ii != iiEnd; ++ii ) {

			Accuracy::Approximate const normSquared( EuclideanNormSquared( ( *ii )->approximatePoint ) );

			int comparison = Accuracy::Approximate::Compare( normSquared, m_innerBoundSquared.GetApproximate() );
			if ( comparison == 0 ) {

				CalculateExact( *ii );
				BOOST_ASSERT( ( *ii )->pExactPoint != NULL );

				mpq_class exactNormSquared( Square( ( *( *ii )->pExactPoint )[ 0 ] ) );
				for ( unsigned int jj = 1; jj < t_Dimension; ++jj )
					exactNormSquared += Square( ( *( *ii )->pExactPoint )[ jj ] );

				if ( exactNormSquared > m_innerBoundSquared.GetExact() )
					comparison = 1;
				else
					comparison = -1;
			}
			BOOST_ASSERT( comparison != 0 );

			if ( comparison > 0 ) {

				pResult = boost::shared_ptr< boost::array< t_Type, t_Dimension > >( new boost::array< t_Type, t_Dimension > );

				{	double const approximateNormSquared = normSquared.GetValue();
					double const approximateInnerBoundSquared =           m_innerBoundSquared.GetApproximate().GetValue()                           / approximateNormSquared;
					double const approximateOuterBoundSquared = std::min( m_outerBoundSquared.GetApproximate().GetValue(), approximateNormSquared ) / approximateNormSquared;

					double const innerScale = std::sqrt( approximateInnerBoundSquared ) * ( 1.0 - lambda );
					double const outerScale = std::sqrt( approximateOuterBoundSquared ) *         lambda;
					double const scale = innerScale + outerScale;

					for ( unsigned int jj = 0; jj < t_Dimension; ++jj )
						( *pResult )[ jj ] = m_center[ jj ] + ( *ii )->approximatePoint[ jj ].GetValue() * scale;
				}

				if ( IsSteiner( *pResult ) ) {

					isSteiner = true;
					break;
				}

				{	CalculateExact( *ii );
					BOOST_ASSERT( ( *ii )->pExactPoint != NULL );

					mpq_class exactNormSquared( Square( ( *( *ii )->pExactPoint )[ 0 ] ) );
					for ( unsigned int jj = 1; jj < t_Dimension; ++jj )
						exactNormSquared += Square( ( *( *ii )->pExactPoint )[ jj ] );

					mpq_class const exactInnerBoundSquared =           m_innerBoundSquared.GetExact()                     / exactNormSquared;
					mpq_class const exactOuterBoundSquared = std::min( m_outerBoundSquared.GetExact(), exactNormSquared ) / exactNormSquared;

					mpq_class const innerScale = mpq_class( std::sqrt( exactInnerBoundSquared.get_d() ) ) * ( mpq_class( 1.0 ) - mpq_class( lambda ) );
					mpq_class const outerScale = mpq_class( std::sqrt( exactOuterBoundSquared.get_d() ) ) *                      mpq_class( lambda );
					mpq_class const scale = innerScale + outerScale;

					for ( unsigned int jj = 0; jj < t_Dimension; ++jj ) {

						mpq_class const result = mpq_class( m_center[ jj ] ) + ( *( *ii )->pExactPoint )[ jj ] * scale;
						( *pResult )[ jj ] = result.get_d();
					}
				}

				/*
					pResult might not now satisfy IsSteiner, due to rounding
					from exact to floating-point, so if it isn't a Steiner
					point, we'll keep looking, but will return what we've got
					if we don't find anything better
				*/
				if ( IsSteiner( *pResult ) ) {

					isSteiner = true;
					break;
				}
			}
		}
	}

	// **NOTE: this exception is technically incorrect (and may perhaps be safely commented out)
	if ( ( pResult != NULL ) && ( ! isSteiner ) )
		throw std::runtime_error( "Unable to find Steiner point" );

	return pResult;
}




#endif    /* __cplusplus */

#endif    /* __CLIPPED_VORONOI_FAST_HH__ */
